unit vkbdhelper;

{
  Force focused control visible when Android Virtual Keyboard showed or hiden
  How to use:
  place vkdbhelper into your project uses section.No more code needed.

  Changes
  =======
  2016.4.15
  + Add support for iOS

  2016.4.14
  + Add support for TMemo

  2016.1.16
  * Remove the timer for fast response(Use Idle message replace)
  * Fix a bug when a focus control reshow virtualkeyboard(Thanks ˮ)
  * Fix a bug when use hide virtual keyboard by use hardware back key(Thanks è)
  * Fix a FMX bug :after virtual keyboard shown,can't use hardware back key to exit app
  2016.1.8
  * Fix a bug when user switch to other app
  2015.7.12
  * Fix space after hide ime and rotate
  * Fix rotate detection

}
interface

implementation

uses classes, sysutils, math, System.Types, System.Messaging, FMX.Types,
  FMX.Controls, System.Rtti,
  FMX.Layouts, FMX.text, FMX.scrollbox, FMX.VirtualKeyboard, FMX.Forms,
  FMX.Platform {$IFDEF QLOG_INCLUDED}, qlog{$ENDIF}, typinfo
{$IFDEF ANDROID}, FMX.Platform.Android, FMX.Helpers.Android,
  FMX.VirtualKeyboard.Android, Androidapi.JNI.GraphicsContentViewText,
  Androidapi.JNI.Embarcadero {$ENDIF}
{$IFDEF IOS}
    , Macapi.Helpers, FMX.Platform.iOS, FMX.VirtualKeyboard.iOS,
  iOSapi.Foundation, iOSapi.UIKit
{$ENDIF}
    ;

type
  TVKStateHandler = class(TComponent)
  protected
    FVKMsgId: Integer; // TVKStateChangeMessage ϢĶID
    FSizeMsgId: Integer; // TSizeChangedMessage ϢĶID
    FIdleMsgId: Integer;
    FLastControl: TControl; // һεScrollBox

    FLastMargin: TPointF; // ԭʼ
    FLastAlign: TAlignLayout; // ԭʼ
    FLastBounds: TRectF; // ԭʼλ
    procedure DoVKVisibleChanged(const Sender: TObject;
      const Msg: System.Messaging.TMessage);
    procedure DoSizeChanged(const Sender: TObject;
      const Msg: System.Messaging.TMessage);
    procedure DoAppIdle(const Sender: TObject;
      const Msg: System.Messaging.TMessage);
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    procedure AdjustCtrl(ACtrl: TControl; AVKBounds, ACtrlBounds: TRectF;
      AVKVisible: Boolean);
    function NeedAdjust(ACtrl: TControl; var ACaretRect: TRectF): Boolean;
  public
    constructor Create(AOwner: TComponent); overload; override;
    destructor Destroy; override;
  end;

var
  VKHandler: TVKStateHandler;
{$IFDEF ANDROID}

function GetVKBounds(var ARect: TRect): Boolean; overload;
var
  ContentRect, TotalRect: JRect;
begin
  ContentRect := TJRect.Create;
  TotalRect := TJRect.Create;
  MainActivity.getWindow.getDecorView.getWindowVisibleDisplayFrame(ContentRect);
  MainActivity.getWindow.getDecorView.getDrawingRect(TotalRect);
  Result := TotalRect.Bottom <> ContentRect.Bottom;
  if Result then
  begin
    ARect.Left := TotalRect.Left;
    ARect.Top := ContentRect.Bottom;
    ARect.Right := TotalRect.Right;
    ARect.Bottom := TotalRect.Bottom;
  end;
end;

function GetVKBounds: TRectF; overload;
var
  ContentRect, TotalRect: JRect;
begin
  ContentRect := TJRect.Create;
  TotalRect := TJRect.Create;
  MainActivity.getWindow.getDecorView.getWindowVisibleDisplayFrame(ContentRect);
  MainActivity.getWindow.getDecorView.getDrawingRect(TotalRect);
  Result := TRectF.Create(ConvertPixelToPoint(TPointF.Create(TotalRect.Left,
    TotalRect.Top + ContentRect.height)),
    ConvertPixelToPoint(TPointF.Create(TotalRect.Right, TotalRect.Bottom)));
end;
{$ELSE}
{$IFDEF IOS}
  _IOS_VKBounds: TRectF;

function GetVKBounds: TRectF; overload;
var
  ATop: Integer;
begin
  Result := _IOS_VKBounds;
  ATop := Screen.WorkAreaTop;
  Result.Top := Result.Top - ATop;
  Result.Bottom := Result.Bottom - ATop;
end;

function GetVKBounds(var ARect: TRect): Boolean; overload;
var
  ATemp: TRectF;
  AService: IFMXScreenService;
  AScale: Single;
begin
  ATemp := GetVKBounds;
  Result := not ATemp.IsEmpty;
  if Result then
  begin
    if TPlatformServices.Current.SupportsPlatformService(IFMXScreenService,
      AService) then
    begin
      AScale := AService.GetScreenScale;
      ARect.Left := Trunc(ATemp.Left * AScale);
      ARect.Top := Trunc(ATemp.Top * AScale);
      ARect.Right := Trunc(ATemp.Right * AScale);
      ARect.Bottom := Trunc(ATemp.Bottom * AScale);
    end;
  end;
end;

{$ENDIF}
{$ENDIF}

/// MainActivityĿͻͼСȷǷʾ
function IsVKVisible: Boolean;
var
  R: TRect;
begin
  Result := GetVKBounds(R);
end;

{ TVKStateHandler }
procedure TVKStateHandler.AdjustCtrl(ACtrl: TControl;
  AVKBounds, ACtrlBounds: TRectF; AVKVisible: Boolean);
var
  ATarget, ACaretRect: TRectF;
  // ƶָϵӶµĸ
  procedure MoveCtrls(AOldParent, ANewParent: TFmxObject);
  var
    I: Integer;
    AChild: TFmxObject;
  begin
    I := 0;
    while I < AOldParent.ChildrenCount do
    begin
      AChild := AOldParent.Children[I];
      if AChild <> ANewParent then
      begin
        if AChild.Parent = AOldParent then
        begin
          AChild.Parent := ANewParent;
          Continue;
        end;
      end;
      Inc(I);
    end;
  end;
  procedure AdjustByLayout(ARoot: TFmxObject; AOffset: Single);
  var
    ALayout: TLayout;
  begin
{$IFDEF QLOG_INCLUDED}
    PostLog(llDebug, '[VKHelper] ǰΪ:%s - %sҪ %f',
      [ARoot.ClassName, ARoot.Name, AOffset]);
    PostLog(llDebug, '[VKHelper] ǰһӶΪ%s - %s',
      [ARoot.Children[0].ClassName, ARoot.Children[0].Name]);
{$ENDIF}
    if (ARoot.ChildrenCount = 1) and (ARoot.Children[0] is TLayout) then
    // 1˵ûLayout
    begin
      ALayout := ARoot.Children[0] as TLayout;
{$IFDEF QLOG_INCLUDED}
      PostLog(llDebug, '[VKHelper] ҵ Layout: %s', [ALayout.Name]);
{$ENDIF}
{$IFDEF QLOG_INCLUDED}
      PostLog(llDebug, '[VKHelper]Layoutؼ %s ǰ', [ALayout.Name]);
      PostLog(llDebug, '[VKHelper] .Align=%s',
        [GetEnumName(TypeInfo(TAlignLayout), Integer(ALayout.Align))]);
      PostLog(llDebug, '[VKHelper] .BoundsRect={%f,%f)-(%f,%f)',
        [ALayout.BoundsRect.Left, ALayout.BoundsRect.Top,
        ALayout.BoundsRect.Right, ALayout.BoundsRect.Bottom]);
      PostLog(llDebug, '[VKHelper] .Postion=(%f,%f)',
        [ALayout.Position.x, ALayout.Position.y]);
{$ENDIF}
      if ALayout.Align <> TAlignLayout.None then
      begin
        FLastAlign := ALayout.Align;
      end;
      FLastBounds := ALayout.BoundsRect;
      FLastMargin.y := ALayout.Position.y;
      FLastMargin.x := ALayout.Position.x;
      ALayout.Align := TAlignLayout.None;
    end
    else
    begin
{$IFDEF QLOG_INCLUDED}
      PostLog(llDebug, '[VKHelper]ûпɵؼһLayout');
{$ENDIF}
      ALayout := TLayout.Create(ARoot);
      ALayout.Parent := ARoot;
      ALayout.TagObject := Self;
      MoveCtrls(ARoot, ALayout);
      FLastMargin.y := 0;
      FLastMargin.x := 0;
      FLastAlign := TAlignLayout.Client;
      FLastBounds := TRectF.Create(0, 0, 0, 0);
    end;
    if ARoot is TForm then
    begin
      if (TForm(ARoot).width <> ALayout.width) or
        (TForm(ARoot).height <> ALayout.height) then
        ALayout.SetBounds(0, 0, TForm(ARoot).width, TForm(ARoot).height);
    end;
    // ûй򣬾ͼһLayoutȻٽеԪƶLayout
    if ALayout.Position.y + AOffset > 0 then
      ALayout.Position.y := 0
    else
      ALayout.Position.y := ALayout.Position.y + AOffset;
{$IFDEF QLOG_INCLUDED}
    PostLog(llDebug, '[VKHelper]');
    PostLog(llDebug, '[VKHelper] .Align=%s',
      [GetEnumName(TypeInfo(TAlignLayout), Integer(ALayout.Align))]);
    PostLog(llDebug, '[VKHelper] .BoundsRect={%f,%f)-(%f,%f)',
      [ALayout.BoundsRect.Left, ALayout.BoundsRect.Top,
      ALayout.BoundsRect.Right, ALayout.BoundsRect.Bottom]);
    PostLog(llDebug, '[VKHelper] .Postion=(%f,%f)',
      [ALayout.Position.x, ALayout.Position.y]);
{$ENDIF}
    if FLastControl <> ALayout then
    begin
      if Assigned(FLastControl) then
        FLastControl.RemoveFreeNotification(Self);
      FLastControl := ALayout;
      FLastControl.FreeNotification(Self);
    end;
  end;
  function AdjustByScrollBox(AScrollBox: TCustomScrollBox): Boolean;
  var
    AParentBounds: TRectF;
    AOffset: Single;
  begin
    AParentBounds := AScrollBox.AbsoluteRect;
    Result := AParentBounds.Contains(ACaretRect);
    if Result then
    begin
      with AScrollBox.ViewportPosition do
        AOffset := y + AVKBounds.Top - ACaretRect.Bottom;
      if (FLastControl <> AScrollBox) then
      begin
        if Assigned(FLastControl) then
          FLastControl.RemoveFreeNotification(Self);
{$IFDEF QLOG_INCLUDED}
        PostLog(llDebug, '[VKHelper] ScrollBox ؼ:%s', [AScrollBox.Name]);
{$ENDIF}
        FLastMargin.y := AScrollBox.Margins.Bottom;
        FLastMargin.x := AScrollBox.Margins.Left;
        FLastControl := AScrollBox;
        FLastControl.FreeNotification(Self);
      end;
      AScrollBox.Margins.Bottom := AParentBounds.Bottom - AVKBounds.Top;
      // ҪһͻݣԱָ֤ܹ
      AScrollBox.ViewportPosition :=
        TPointF.Create(AScrollBox.ViewportPosition.x,
        AScrollBox.ViewportPosition.y + AOffset);
    end;
  end;

  function AdjustByPresentedScrollBox(AScrollBox
    : TCustomPresentedScrollBox): Boolean;
  var
    AParentBounds: TRectF;
    ALast, AOffset: TPointF;
  begin
    AParentBounds := AScrollBox.AbsoluteRect;
    Result := AParentBounds.Contains(ACaretRect);
    if Result then
    begin
      with AScrollBox.ViewportPosition do
      begin
        AOffset.x := x;
        if ACaretRect.Bottom < AVKBounds.Top then
          AOffset.y := y - ACaretRect.Top
        else
          AOffset.y := y + AVKBounds.Top - ACaretRect.Bottom;
      end;
      // ҪһͻݣԱָ֤ܹ
      ALast := AScrollBox.ViewportPosition;
      AScrollBox.ScrollTo(AOffset.x, AOffset.y);
      Result := (SameValue(ALast.x - AScrollBox.ViewportPosition.x, AOffset.x)
        and SameValue(ALast.y - AScrollBox.ViewportPosition.y, AOffset.y));
      if Result then
      begin
        if (FLastControl <> AScrollBox) then
        begin
          if Assigned(FLastControl) then
            FLastControl.RemoveFreeNotification(Self);
{$IFDEF QLOG_INCLUDED}
          PostLog(llDebug, '[VKHelper] PresentedScrollBox ؼ:%s',
            [AScrollBox.Name]);
{$ENDIF}
          FLastMargin := AScrollBox.ViewportPosition;
          FLastControl := AScrollBox;
          FLastControl.FreeNotification(Self);
        end;
        AScrollBox.Margins.Bottom := AParentBounds.Bottom - AVKBounds.Top;
{$IFDEF QLOG_INCLUDED}
        PostLog(llDebug, '[VKHelper]ƫ:(%f,%f),ǰƫƣ(%f,%f)',
          [AOffset.x, AOffset.y, AScrollBox.ViewportPosition.x,
          AScrollBox.ViewportPosition.y]);
{$ENDIF}
      end
      else
      begin
        with AScrollBox.ViewportPosition do
          OffsetRect(ACaretRect, ALast.x - x, ALast.y - y);
        AScrollBox.ViewportPosition := ALast;
      end;
    end;
  end;
/// ָƶ
  procedure ScrollInToRect;
  var
    AParent, ALastParent: TFmxObject;
    AParentBounds: TRectF;
    AOffset: Single;
    AHasScrollBox: Boolean;
  begin
    AParent := ACtrl;
    ALastParent := AParent.Parent;
    AHasScrollBox := False;
    while Assigned(AParent) do
    begin
      if AParent is TCustomScrollBox then
      // йܹܲ
      begin
        AHasScrollBox := True;
        if AdjustByScrollBox(AParent as TCustomScrollBox) then
          Exit;
      end
      else if AParent is TCustomPresentedScrollBox then
      begin
        AHasScrollBox := True;
        if AdjustByPresentedScrollBox(AParent as TCustomPresentedScrollBox) then
          Exit;
      end;
      ALastParent := AParent;
      AParent := AParent.Parent;
    end;
    if AHasScrollBox then
    begin
      if ACaretRect.Bottom < AVKBounds.Top then
        AdjustByLayout(ALastParent, Screen.WorkAreaTop - ACaretRect.Top)
      else
        AdjustByLayout(ALastParent, AVKBounds.Top - ACaretRect.Bottom);
    end
    else
    begin
      if ACaretRect.Bottom < AVKBounds.Top then
        AdjustByLayout(ALastParent, ACaretRect.Bottom - ACtrlBounds.Top)
      else
        AdjustByLayout(ALastParent, AVKBounds.Top - ACtrlBounds.Bottom);
    end;
  end;

  function ComparePos(V1, V2: Single): Integer;
  var
    ADelta: Single;
  begin
    ADelta := V1 - V2;
    if ADelta > 0.0001 then
      Result := 1
    else if ADelta < -0.0001 then
      Result := -1
    else
      Result := 0;
  end;

begin
  if AVKVisible then
  begin
    if NeedAdjust(ACtrl, ACaretRect) then
    begin
{$IFDEF QLOG_INCLUDED}
      PostLog(llDebug,
        '[VKHelper]ؼ %s(%s) :(%f,%f)-(%f,%f),:(%f,%f)-(%f,%f)',
        [ACtrl.Name, ACtrl.ClassName, ACtrlBounds.Left, ACtrlBounds.Top,
        ACtrlBounds.Right, ACtrlBounds.Bottom, AVKBounds.Left, AVKBounds.Top,
        AVKBounds.Right, AVKBounds.Bottom]);
{$ENDIF}
      ScrollInToRect;
    end
    else
    begin
      // {$IFDEF QLOG_INCLUDED}
      // PostLog(llHint, '[VKHelper]ʾǰƿɼҪ');
      // {$ENDIF}
    end
  end
  else
  begin
    if Assigned(FLastControl) then
    begin
      if (FLastControl is TCustomScrollBox) or
        (FLastControl is TCustomPresentedScrollBox) then
      begin
        FLastControl.Margins.Bottom := FLastMargin.y;
        FLastControl.Margins.Left := FLastMargin.x;
      end
      else
      begin
        if FLastAlign = TAlignLayout.None then
        begin
          FLastControl.Position.y := FLastMargin.y;
          FLastControl.Position.x := FLastMargin.x;
        end
        else
        begin
          FLastControl.BoundsRect := FLastBounds;
          FLastControl.Align := FLastAlign;
        end;
      end;
      FLastControl := nil;
    end;
  end;
end;

// 캯Ϣ
constructor TVKStateHandler.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FVKMsgId := TMessageManager.DefaultManager.SubscribeToMessage
    (TVKStateChangeMessage, DoVKVisibleChanged);
  FSizeMsgId := TMessageManager.DefaultManager.SubscribeToMessage
    (TSizeChangedMessage, DoSizeChanged);
  FIdleMsgId := TMessageManager.DefaultManager.SubscribeToMessage(TIdleMessage,
    DoAppIdle);
end;

/// ȡϢ
destructor TVKStateHandler.Destroy;
begin
  TMessageManager.DefaultManager.Unsubscribe(TVKStateChangeMessage, FVKMsgId);
  TMessageManager.DefaultManager.Unsubscribe(TSizeChangedMessage, FSizeMsgId);
  TMessageManager.DefaultManager.Unsubscribe(TIdleMessage, FIdleMsgId);
  inherited;
end;

/// ӦÿʱǷػǷ񸲸ס˵ǰýĿؼ
procedure TVKStateHandler.DoAppIdle(const Sender: TObject;
  const Msg: System.Messaging.TMessage);
{$IFDEF ANDROID}
  procedure FMXAndroidFix;
  var
    AService: IFMXVirtualKeyboardService;
    AVK: TVirtualKeyboardAndroid;
    AListener: TVKListener;
    AContext: TRttiContext;
    AType: TRttiType;
    AField: TRttiField;
  begin
    if TPlatformServices.Current.SupportsPlatformService
      (IFMXVirtualKeyboardService, AService) then
    begin
      AVK := AService as TVirtualKeyboardAndroid;
      AContext := TRttiContext.Create;
      AType := AContext.GetType(AVK.ClassType);
      if Assigned(AType) then
      begin
        AField := AType.GetField('FVKListener');
        if Assigned(AField) then
        begin
          AListener := AField.GetValue(AVK).AsObject as TVKListener;
          AListener.onVirtualKeyboardHidden;
        end;
      end;
      Screen.ActiveForm.Focused := nil;
    end;
  end;

  procedure UpdateAndroidKeyboardServiceState;
  var
    ASvc: IFMXVirtualKeyboardService;
    AContext: TRttiContext;
    AType: TRttiType;
    AField: TRttiField;
    AInst: TVirtualKeyboardAndroid;
  begin
    if (not Assigned(Screen.FocusControl)) and
      TPlatformServices.Current.SupportsPlatformService
      (IFMXVirtualKeyboardService, ASvc) then
    begin
      AInst := ASvc as TVirtualKeyboardAndroid;
      AContext := TRttiContext.Create;
      AType := AContext.GetType(TVirtualKeyboardAndroid);
      AField := AType.GetField('FState');
      if AField.GetValue(AInst).AsOrdinal <> 0 then
        PByte(AInst)[AField.Offset] := 0;
    end;
  end;
{$ENDIF}
  procedure CheckHidden;
  var
    ACaretRect: TRectF;
    ACtrl: TControl;
  begin
    ACtrl := Screen.FocusControl.GetObject as TControl;
    if Assigned(ACtrl) then
    begin
      if NeedAdjust(ACtrl, ACaretRect) then
        AdjustCtrl(ACtrl, GetVKBounds, ACtrl.AbsoluteRect, True);
    end;
  end;

begin
  if not IsVKVisible then // غ
  begin
    if Assigned(FLastControl) then
    begin
      TMessageManager.DefaultManager.SendMessage(FLastControl,
        TVKStateChangeMessage.Create(False, TRect.Create(0, 0, 0, 0)));
    end
{$IFDEF ANDROID}
    else
      UpdateAndroidKeyboardServiceState;
{$ENDIF}
  end
  else if Assigned(Screen.FocusControl) then
    CheckHidden;
end;

/// ںлʱؼλ
procedure TVKStateHandler.DoSizeChanged(const Sender: TObject;
  const Msg: System.Messaging.TMessage);
var
  ASizeMsg: TSizeChangedMessage absolute Msg;
  R: TRect;
  AScene: IScene;
  AScale: Single;
begin
  if Sender = Screen.ActiveForm then
  begin
    if GetVKBounds(R) then
    begin
      if Assigned(FLastControl) then
      begin
        if FLastControl is TLayout then
        begin
          if FLastControl.TagObject = Self then
          begin
{$IFDEF QLOG_INCLUDED}
            PostLog(llHint, '[VKHelper] Layoutɱ󴴽ΪĻߴ');
{$ENDIF}
            FLastControl.SetBounds(0, 0, Screen.ActiveForm.width,
              Screen.ActiveForm.height);
          end;
        end
        else //
          TCustomScrollBox(FLastControl).Margins.Bottom := 0;
      end;
      if Supports(Sender, IScene, AScene) then
      begin
        AScale := AScene.GetSceneScale;
        R.Left := Trunc(R.Left / AScale);
        R.Top := Trunc(R.Top / AScale);
        R.Right := Trunc(R.Right / AScale);
        R.Bottom := Trunc(R.Bottom / AScale);
{$IFDEF QLOG_INCLUDED}
        PostLog(llHint, '[VKHelper]ĻߴΪ %d - %d(%d,%d)-(%d-%d)',
          [ASizeMsg.Value.width, ASizeMsg.Value.height, R.Left, R.Top, R.Right,
          R.Bottom]);
{$ENDIF}
        TMessageManager.DefaultManager.SendMessage(Sender,
          TVKStateChangeMessage.Create(True, R));
      end;
    end
{$IFDEF QLOG_INCLUDED}
    else
      PostLog(llHint, '[VKHelper]ĻߴΪ %d - %d̴״̬',
        [ASizeMsg.Value.width, ASizeMsg.Value.height]);
{$ENDIF}
  end;
end;

/// ̿ɼԱϢָؼλ
procedure TVKStateHandler.DoVKVisibleChanged(const Sender: TObject;
  const Msg: System.Messaging.TMessage);
var
  AVKMsg: TVKStateChangeMessage absolute Msg;
  ACtrl: TControl;
begin
  if AVKMsg.KeyboardVisible then // ̿ɼ
  begin
{$IFDEF IOS}
    _IOS_VKBounds := TRectF.Create(AVKMsg.KeyboardBounds);
{$ENDIF}
    if Screen.FocusControl <> nil then
    begin
      ACtrl := Screen.FocusControl.GetObject as TControl;
      AdjustCtrl(ACtrl, GetVKBounds, ACtrl.AbsoluteRect, True);
    end
{$IFDEF QLOG_INCLUDED}
    else
      PostLog(llHint, '[VKHelper]ʾScreen.FocusControlȴΪ');
{$ENDIF}
  end
  else
  begin
{$IFDEF IOS}
    _IOS_VKBounds := TRectF.Empty;
{$ENDIF}
    if Assigned(FLastControl) then // 
      AdjustCtrl(FLastControl, GetVKBounds, FLastControl.AbsoluteRect, False);
  end;
end;

/// Ӧͷ֪ͨԱЧַ
function TVKStateHandler.NeedAdjust(ACtrl: TControl;
  var ACaretRect: TRectF): Boolean;
var
  ACaret: ICaret;
  ACaretObj: TCustomCaret;
  ACtrlBounds, AVKBounds: TRectF;
  AWorkArea: TRect;
  S: String;
  function GetRootTop: Single;
  var
    ALayout: TLayout;
    AObj: TFmxObject;
  begin
    AObj := ACtrl.Root.GetObject;
    Result := 0;
    if AObj is TForm then
    begin
      if TForm(AObj).ChildrenCount > 0 then
      begin
        AObj := TForm(AObj).Children[0];
        if AObj is TLayout then
        begin
          ALayout := AObj as TLayout;
          Result := ALayout.Position.y;
        end;
      end;
    end;
  end;

begin
  if Supports(ACtrl, ICaret, ACaret) then
  begin
    AVKBounds := GetVKBounds;
    ACtrlBounds := ACtrl.AbsoluteRect;
    ACaretObj := ACaret.GetObject;
    ACaretRect.TopLeft := ACtrl.LocalToAbsolute(ACaretObj.Pos);
    ACaretRect.Right := ACaretRect.Left + ACaretObj.size.cx;
    ACaretRect.Bottom := ACaretRect.Top + ACaretObj.size.cy + 5; // ӵ
    // AWorkArea := Screen.WorkAreaRect;
    // FMX.Memo Ȼ⣬ʱ
    Result := ACaretRect.IntersectsWith(AVKBounds) or (ACaretRect.Top < 0);
    if Result then
    begin
{$IFDEF QLOG_INCLUDED}
      PostLog(llDebug,
        '꣺(%f,%f)-(%f,%f),ؼ(%f,%f)-(%f,%f)̣(%f,%f)-(%f,%f)',
        [ACaretRect.Left, ACaretRect.Top, ACaretRect.Right, ACaretRect.Bottom,
        ACtrlBounds.Left, ACtrlBounds.Top, ACtrlBounds.Right,
        ACtrlBounds.Bottom, AVKBounds.Left, AVKBounds.Top, AVKBounds.Right,
        AVKBounds.Bottom]);
{$ENDIF}
    end;
  end;
end;

procedure TVKStateHandler.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  if Operation = opRemove then
  begin
    if FLastControl = AComponent then
      FLastControl := nil;
  end;
  inherited;
end;

initialization

// ֧Android+IOS
{$IF DEFINED(ANDROID) OR DEFINED(IOS)}
  VKHandler := TVKStateHandler.Create(nil);
{$ENDIF}

finalization

{$IF DEFINED(ANDROID)  OR DEFINED(IOS)}
  FreeAndNil(VKHandler);
{$ENDIF}

end.
